/* ``The contents of this file are subject to the Erlang Public License,
 * Version 1.1, (the "License"); you may not use this file except in
 * compliance with the License. You should have received a copy of the
 * Erlang Public License along with this software. If not, it can be
 * retrieved via the world wide web at http://www.erlang.org/.
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Original Code is Ericsson Utvecklings AB.
 * Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
 * AB. All Rights Reserved.''
 * 
 *     $Id$
 */
/*
 * Function:
 * ei_print_term to print out a binary coded term
 */

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#ifdef VRTX
#define __READY_EXTENSIONS__
#include <errno.h>
#endif

#include "ei.h"
#include "ei_printterm.h"
#include "erl_error.h"
#include "erl_malloc.h"


#define BINPRINTSIZE 30

/*
 * PRINT out a binary term (hacked from erl_print_term)
 */

static int print_string(FILE* fp, char* s);

int ei_print_term(FILE *fp, const char* buf, int* index)
{
    int i, doquote, n, m, ty;
    char a[MAXATOMLEN], *p;
    int ch_written = 0;		/* counter of written chars */
    erlang_pid pid;
    erlang_port port;
    erlang_ref ref;
    double d;
    long l;
    unsigned long u;

    if (fp == NULL) return -1;
    /* ASSERT(ep != NULL); */

    doquote = 0;
    ei_get_type(buf, index, &ty, &n);
    switch (ty) {
    case ERL_ATOM_EXT:
	/* XXX: what if some weird locale is in use? */
	if (ei_decode_atom(buf, index, a) < 0)
	   goto err;
	doquote = !islower(a[0]);
	for (p = a; !doquote && *p != '\0'; ++p)
	    doquote = !(isalnum(*p) || *p == '_' || *p == '@');
	if (doquote) {
	    putc('\'', fp); ++ch_written; 
	}
	fputs(a, fp); ch_written += strlen(a);
	if (doquote) {
	    putc('\'', fp); ++ch_written;
	}
	break;
    /*case ERL_VARIABLE:
        if (!isupper(ERL_VAR_NAME(ep)[0])) {
	    doquote = 1;
	    putc('\'', fp);
	    ch_written++;
	}
	      
	fputs(ERL_VAR_NAME(ep), fp);
	ch_written += ERL_VAR_LEN(ep);
	      
	if (doquote) {
	    putc('\'', fp);
	    ch_written++;
	}
	break;*/
    case ERL_PID_EXT:
	if (ei_decode_pid(buf, index, &pid) < 0) goto err;
	ch_written += fprintf(fp, "<%s.%d.%d>", pid.node, pid.num, pid.serial);
	break;
    case ERL_PORT_EXT:
	if (ei_decode_port(buf, index, &port) < 0) goto err;
	ch_written += fprintf(fp, "#Port<%d.%d>", port.id, port.creation); /* eller ska id delas upp i tv <id'.id''>? */
	break;
    case ERL_NEW_REFERENCE_EXT:
    case ERL_REFERENCE_EXT:
	if (ei_decode_ref(buf, index, &ref) < 0) goto err;
	ch_written += fprintf(fp, "#Ref<");
	for (i = 0; i < ref.len; ++i) {
	    ch_written += fprintf(fp, "%d", ref.n[i]);
	    if (i < ref.len - 1) {
		putc('.', fp); ++ch_written;
	    }
	}
	putc('>', fp); ++ch_written;
	break;
    case ERL_NIL_EXT:
	if (ei_decode_list_header(buf, index, &n) < 0) goto err;
	ch_written += fprintf(fp, "[]");
	break;
    case ERL_LIST_EXT:
	if (ei_decode_list_header(buf, index, &n) < 0) goto err;
	putc('[', fp); ch_written++;
	for (i = 0; i < n; ++i) {
	    ch_written += ei_print_term(fp, buf, index);
	    if (i < n - 1) {
		fputs(", ", fp); ch_written += 2;
	    }
	}
	if (ei_get_type(buf, index, &ty, &n) < 0) goto err;
	if (ty != ERL_NIL_EXT) {
	    fputs(" | ", fp); ch_written += 3;
	    ch_written += ei_print_term(fp, buf, index);
	} else {
	    if (ei_decode_list_header(buf, index, &n) < 0) goto err;
	}
	putc(']', fp); ch_written++;
	break;
    case ERL_STRING_EXT:
	p = erl_malloc(n+1);
	if (p == NULL) goto err;
	if (ei_decode_string(buf, index, p) < 0) {
	    erl_free(p);
	    goto err;
	}
	ch_written += print_string(fp, p);
	erl_free(p);
	break;
    case ERL_SMALL_TUPLE_EXT:
    case ERL_LARGE_TUPLE_EXT:
	if (ei_decode_tuple_header(buf, index, &n) < 0) goto err;	
	putc('{', fp); ch_written++;
	for (i = 0; i < n; ++i) {
	    ch_written += ei_print_term(fp, buf, index);
	    if (i < n-1) {
		fputs(", ", fp); ch_written += 2;
	    }
	}
	putc('}', fp); ch_written++;
	break;
    case ERL_BINARY_EXT:
	p = erl_malloc(n);
	if (p == NULL) goto err;
	if (ei_decode_binary(buf, index, p, (long*)&n) < 0) {
	    erl_free(p);
	    goto err;
	}
	ch_written += fprintf(fp, "#Bin<");
	m = n;
	if (n > BINPRINTSIZE)
	    m = BINPRINTSIZE;
	--m;
	for (i = 0; i < m; ++i) {
	    ch_written += fprintf(fp, "%d,", p[i]);
	}
	ch_written += fprintf(fp, "%d", p[i]);
	if (n > BINPRINTSIZE)
	    ch_written += fprintf(fp, ",...");
	putc('>', fp); ++ch_written;
	erl_free(p);
	break;
    case ERL_SMALL_INTEGER_EXT:
    case ERL_INTEGER_EXT:
	if (ei_decode_long(buf, index, &l) < 0) goto err;
	ch_written += fprintf(fp, "%ld", l);
	break;
    case ERL_SMALL_BIG_EXT:
	if (ei_decode_ulong(buf, index, &u) < 0) goto err;
	ch_written += fprintf(fp, "%lu", u);
	break;
    case ERL_LARGE_BIG_EXT:
	ch_written += fprintf(fp, "BigNum");
	*index += n;
	break;
    case ERL_FLOAT_EXT:
	if (ei_decode_double(buf, index, &d) < 0) goto err;
	ch_written += fprintf(fp, "%f", d);
	break;
    default:
	goto err;
    }
    return ch_written;
 err:
    erl_err_msg("<ERROR> ei_print_term: Bad type of term !");
    return -1;
}

static int print_string(FILE* fp, char* s)
{
    int ch_written = 0;		/* counter of written chars */
  
    putc('"', fp);
    ++ch_written;
    for (; *s != '\0'; ++s) {
	int c = *s;
	if (c >= ' ') {
	    putc(c, fp); ++ch_written; }
	else {
	    switch (c) {
	    case '\n': fputs("\\n", fp); ch_written += 2; break;
	    case '\r': fputs("\\r", fp); ch_written += 2; break;
	    case '\t': fputs("\\t", fp); ch_written += 2; break;
	    case '\v': fputs("\\v", fp); ch_written += 2; break;
	    case '\b': fputs("\\b", fp); ch_written += 2; break;
	    case '\f': fputs("\\f", fp); ch_written += 2; break;
		break;
	    default:
		ch_written += fprintf(fp, "\\x%x", c);
		break;
	    }
	}
    }
    putc('"', fp); ch_written++;
    return ch_written;
}


/* ------------------------------------------ */

/*
 * skip a binary term
 */

static int print_string(FILE* fp, char* s);

int ei_skip_term(const char* buf, int* index)
{
    int i, n, ty;

    /* ASSERT(ep != NULL); */

    ei_get_type(buf, index, &ty, &n);
    switch (ty) {
    case ERL_ATOM_EXT:
	/* XXX: what if some weird locale is in use? */
	if (ei_decode_atom(buf, index, NULL) < 0)
	   goto err;
	break;
    case ERL_PID_EXT:
	if (ei_decode_pid(buf, index, NULL) < 0) goto err;
	break;
    case ERL_PORT_EXT:
	if (ei_decode_port(buf, index, NULL) < 0) goto err;
	break;
    case ERL_NEW_REFERENCE_EXT:
    case ERL_REFERENCE_EXT:
	if (ei_decode_ref(buf, index, NULL) < 0) goto err;
	break;
    case ERL_NIL_EXT:
	if (ei_decode_list_header(buf, index, &n) < 0) goto err;
	break;
    case ERL_LIST_EXT:
	if (ei_decode_list_header(buf, index, &n) < 0) goto err;
	for (i = 0; i < n; ++i)
	    ei_skip_term(buf, index);
	if (ei_get_type(buf, index, &ty, &n) < 0) goto err;
	if (ty != ERL_NIL_EXT)
	    ei_skip_term(buf, index);
	else
	    if (ei_decode_list_header(buf, index, &n) < 0) goto err;
	break;
    case ERL_STRING_EXT:
	if (ei_decode_string(buf, index, NULL) < 0) goto err;
	break;
    case ERL_SMALL_TUPLE_EXT:
    case ERL_LARGE_TUPLE_EXT:
	if (ei_decode_tuple_header(buf, index, &n) < 0) goto err;	
	for (i = 0; i < n; ++i)
	    ei_skip_term(buf, index);
	break;
    case ERL_BINARY_EXT:
	if (ei_decode_binary(buf, index, NULL, (long*)&n) < 0)
	    goto err;
	break;
    case ERL_SMALL_INTEGER_EXT:
    case ERL_INTEGER_EXT:
	if (ei_decode_long(buf, index, NULL) < 0) goto err;
	break;
    case ERL_SMALL_BIG_EXT:
	if (ei_decode_ulong(buf, index, NULL) < 0) goto err;
	break;
    case ERL_LARGE_BIG_EXT:
	*index += n; /* !! Funkar detta? */
	break;
    case ERL_FLOAT_EXT:
	if (ei_decode_double(buf, index, NULL) < 0) goto err;
	break;
    case ERL_FUN_EXT:
	if (ei_decode_fun(buf, index, NULL) < 0) goto err;
	break;
    default:
	goto err;
    }
    return 0;
 err:
    erl_err_msg("<ERROR> ei_skip_term: unknown type of term !");
    return -1;
}
