%%%----------------------------------------------------------------------
%%% File    : hipe_color.erl
%%% Author  : Ingemar berg <d95ina@zeppo.it.uu.se>
%%% Purpose : Provide a datatype for a color mapping, mapping registers
%%%           to colors.
%%% Created : 19 Mar 2000 by Ingemar berg <d95ina@zeppo.it.uu.se>
%%%----------------------------------------------------------------------

-module(hipe_color).
-author('d95ina@zeppo.it.uu.se').
-export([init/1, getColor/2, setColor/3]).

%%%----------------------------------------------------------------------
% Function: init
%
% Description:  Initialize a color mapping with NrOfNodes slots.
% Parameters:
%   NrOfNodes      -- The maximum number of the nodes that shall be colored.
%
% Returns:
%   Color          -- The brand new color mapping.
%
%%%----------------------------------------------------------------------

init(NrOfNodes) ->
    {color, hipe_vectors:empty(NrOfNodes,undef)}.

%%%----------------------------------------------------------------------
% Function: getColor
%
% Description:  Get the color of a node in a certain color mapping
% Parameters:
%   Node          -- The node whose color shall be found
%   Color         -- The color mapping we shall find the color in
%
% Returns:
%   Col           -- The color of the node
%
%%%----------------------------------------------------------------------

getColor(Node,{color, Color}) when integer(Node) ->
    hipe_vectors:get(Color,Node);
getColor(Node,{color, Color}) ->
    hipe_message:error(getColor,"Node is not an integer ~p",[Node]).

%%%----------------------------------------------------------------------
% Function: setColor
%
% Description: Set the color of a node in a certain color mapping.
% Parameters: 
%   Node          -- The node to color
%   NodeColor     -- The color the node shall have
%   Color         -- The colormapping the info shall be stored in
% Returns:
%   NewColor      -- The new mapping.
%
%%%----------------------------------------------------------------------

setColor(Node,NodeColor,{color, Color}) when integer(Node) ->
    {color, hipe_vectors:set(Color,Node,NodeColor)};
setColor(Node,NodeColor,{color, Color}) ->
    hipe_message:error(setColor,"Node is not an integer ~p",[Node]).

%%% Return the name of a color, if no name is defined, return its number. 
%%% (This function is used mainly for debugging purposes)

% name(0) -> red;
% name(1) -> orange;
% name(2) -> yellow;
% name(3) -> green;
% name(4) -> blue;
% name(5) -> indigo;
% name(6) -> purple;
% name(7) -> octarine;                                   % The Color Of Magic
% name(X) -> X.
