%%%----------------------------------------------------------------------
%%% File    : hipe_message.erl
%%% Author  : Andreas Wallin <d96awa@zeppo.it.uu.se>
%%% Purpose : Generates debug, warning and error messages.
%%% Created : 5 Mar 2000 by Andreas Wallin <d96awa@zeppo.it.uu.se>
%%%----------------------------------------------------------------------

-module(hipe_message).
-author('d96awa@zeppo.it.uu.se').

-export([debug_enabled/1, debug/2, debug/3,
	 warning/2, warning/3, 
	 error/2, error/3]).

% A silly method for turning debug messages on and off.
% 
debug_enabled(adj_list) ->
    false;
debug_enabled(adj_set) ->
    false;
debug_enabled(alias) ->
    false;
debug_enabled(assign) ->
    false;
debug_enabled(coalesce) ->
    false;
debug_enabled(color) ->
    false;
debug_enabled(degree) ->
    false;
debug_enabled(freeze) ->
    false;
debug_enabled(ig) ->
    false;
debug_enabled(node_sets) ->
    false;
debug_enabled(select) ->
    false;
debug_enabled(stack) ->
    false;
debug_enabled(worklists) ->
    false;
debug_enabled(regalloc) ->
    false;
debug_enabled(kul) ->
    false;
debug_enabled(_) ->
    false.


%%%----------------------------------------------------------------------
% Function:    debug
%
% Description: Generates a debug message
% Parameters:
%   Module  --  The name of the module that the debug message occured in
%   Msg     --  Debug message
%
% Returns:      
%   Nothing
%%%----------------------------------------------------------------------

debug(Module, Msg) ->
    case debug_enabled(Module) of
	true -> 
	    io:format('DEBUG (~p): ~p~n', [Module, Msg]);
	false ->
	    io:format('', [])
    end.

%%%----------------------------------------------------------------------
% Function:    debug
%
% Description: Generates a debug message
% Parameters:
%   Module  --  The name of the module that the debug message occured in
%   Msg     --  Debug message
%   Vs      --  A list of variables referred in the Msg string
%
% Returns:      
%   Nothing
%%%----------------------------------------------------------------------

debug(Module, Msg, Vs) ->
    case debug_enabled(Module) of
	true -> 
	    io:format('DEBUG (~p): ',[Module]),
	    io:format(Msg, Vs),
	    io:format('~n',[]);
	false ->
	    io:format('', [])
    end.



%%%----------------------------------------------------------------------
% Function:    warning
%
% Description: Generates a warning message
% Parameters:
%   Fun     --  The name of the function that the warning occured in
%   Msg     --  Warning message
%
% Returns:      
%   Nothing
%%%----------------------------------------------------------------------
warning(Fun,Msg) ->
   io:format('WARNING (~p): ~p~n',[Fun,Msg]).

%%%----------------------------------------------------------------------
% Function:    warning
%
% Description: Generates a warning message
% Parameters:
%   Fun     --  The name of the function that the warning occured in
%   Msg     --  Warning message that contains io format strings
%   Vs      --  A list of variables referred in the Msg string
%
% Returns:      
%   Nothing
%%%----------------------------------------------------------------------
warning(Fun, Msg, Vs) ->
   io:format('WARNING (~p):',[Fun]),
   io:format(Msg, Vs),
   io:format('~n',[]).

%%%----------------------------------------------------------------------
% Function:    error
%
% Description: Generates an error message and exit
% Parameters:
%   Fun     --  The name of the function that the error occured in
%   Msg     --  Error message
%
% Returns:      
%   Nothing
%%%----------------------------------------------------------------------
error(Fun,Msg) ->
   io:format('ERROR (~p): ~p~n',[Fun,Msg]),
   exit({error,Fun,Msg}).

%%%----------------------------------------------------------------------
% Function:    error
%
% Description: Generates an error message and exit
% Parameters:
%   Fun     --  The name of the function that the error occured in
%   Msg     --  Error message that contains io format strings
%   Vs      --  A list of variables referred in the Msg string
%
% Returns:
%   Nothing
%%%----------------------------------------------------------------------
error(Fun,Msg,Vs) ->
   io:format('ERROR (~p): ',[Fun]),
   io:format(Msg,Vs),
   io:format('~n',[]),
   exit({error,Fun,Msg}).

