%%% $Id: hipe_x86_specific.erl,v 1.4 2001/06/19 16:56:22 ulma4761 Exp $
%%% hipe_x86_specific
%%% This module defines interface to the x86 backend
%%% Copyright (C) Ulf Magnusson
%%% Email: ulf.magnusson@ubm-computing.com

-module(hipe_x86_specific).

-export([analyze/1,
	 liveout/2,
	 allocatable/0,
	 all_precolored/0,
	 is_precolored/1,
	 physical_name/1,
	 labels/1,
	 var_range/1,
	 bb/2,
	 def_use/1,
	 is_move/1,
	 reg_nr/1]).

% Liveness stuff

analyze(CFG) ->
    hipe_x86_liveness:analyze(CFG).

liveout(BB_in_out_liveness,Label) ->
    hipe_x86_liveness:liveout(BB_in_out_liveness,Label).

% Registers stuff

allocatable() ->
    A = hipe_x86_registers:allocatable(),
    lists:map( fun(X) -> X+1 end, A ).

all_precolored() ->
    A = hipe_x86_registers:all_precoloured(),
    lists:map( fun(X) -> X+1 end, A ).

is_precolored(Reg) ->
    hipe_x86_registers:is_precoloured(Reg-1).

physical_name(Reg) ->
    hipe_x86_registers:reg_name(Reg-1).

% CFG stuff

labels(CFG) ->
    hipe_x86_cfg:labels(CFG).

var_range(CFG) ->
    %Linear = hipe_x86_cfg:linearise(CFG),
%     Code = hipe_x86:defun_code(Linear),
%     HV = highest_var(Code),
%     io:format("Var_range:{1,~w}\n",[HV]),
%     {1,HV}.
    %hipe_x86_pp:pp(Linear),
    {Min,Max} = hipe_x86_cfg:var_range(CFG),
    %io:format("Var_range: ~w\n",[{Min,Max}]),
    {Min,Max+1}.
 
bb(CFG,L) ->
    hipe_x86_cfg:bb(CFG,L).

% X86 stuff

def_use(Instruction) ->
    {hipe_x86_defuse:insn_def(Instruction),
     hipe_x86_defuse:insn_use(Instruction)}.

is_move(Instruction) ->
    hipe_x86:is_move(Instruction).

reg_nr(Reg) ->
    %io:format("Reg:~w Lying that it is +1\n",[Reg]),
    hipe_x86:temp_reg(Reg)+1.


