%% Copyright (c) 1998 by Erik Johansson.  All Rights Reserved 
%% Time-stamp: <01/02/14 14:25:59 happi>
%% ====================================================================
%%  Filename : 	data_alloc.erl
%%  Module   :	data_alloc
%%  Purpose  :  
%%  Notes    : 
%%  History  :	* 1998-04-16 Erik Johansson (happi@csd.uu.se): Created.
%% CVS:
%%    $Author: richardc $
%%    $Date: 2001/03/26 18:37:08 $
%%    $Revision: 1.1.1.1 $
%% ====================================================================
%% Exported functions (short description):
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(hipe_data_alloc).
-export([size_of/1,alloc/2,alloc_constants/2,find_address/2]).

size_of(Constant) ->
  case hipe_consttab:const_type(Constant) of
    term -> hipe_bifs:term_size(hipe_consttab:const_data(Constant))*4;
    block -> hipe_consttab:block_size(Constant)
  end.

alloc(Constant,MemoryZone) ->
  case hipe_consttab:const_align(Constant) of
    2 -> true;
    4 -> true;
    8 -> exit("Don't handle dword alignment yet!")
  end,
  Size = size_of(Constant),
  Need = (Size+3) div 4,
  Mem = case hipe_consttab:const_type(Constant) of
	  term -> hipe_bifs:alloc_constant(Need);
	  block -> hipe_bifs:alloc_code(Need)
	end,
  Adr = case hipe_consttab:const_type(Constant) of
	  term ->
	    hipe_bifs:copy_term(hipe_consttab:const_data(Constant),Mem,Need);
	  block ->
	    write_bytes(hipe_consttab:const_data(Constant),
			Mem),
	    Mem
	end,
  %% io:format("Constant at:~w, in block: ~w~n",[Adr,Mem]),
  {Adr,Mem,Size}.

write_bytes([B|Rest],Addr) ->
  hipe_bifs:write_u8(Addr, B),
  write_bytes(Rest,Addr+1);
write_bytes([],_) -> true.


alloc_constants(ConstTab,MFA) ->
  lists:map(fun (Label) ->
		Const = hipe_consttab:lookup(Label,ConstTab),
		{Adr,Mem,Size} = alloc(Const,data), 
		case hipe_consttab:const_exported(Const) of
		  true -> true; %% TODO: Fix this in new system...
						% fix type 
		  false -> true
		end,
		
		{Label,
		 Adr,Mem,
		 hipe_consttab:const_exported(Const)}
	    end,
	    hipe_consttab:labels(ConstTab)).
  
  
find_address(Name,[]) -> exit({no_such_constant,Name});
find_address(Name,[{Label,Address,_,_}|Rest]) ->
  if Name =:= Label -> Address;
     true -> find_address(Name,Rest)
  end.

