%% Copyright (c) 1998 by Erik Johansson.  All Rights Reserved 
%% Time-stamp: <99/12/16 16:12:51 happi>
%% ====================================================================
%%  Filename : 	hipe_link_error.erl
%%  Module   :	hipe_link_error
%%  Purpose  :  
%%  Notes    : 
%%  History  :	* 1998-06-18 Erik Johansson (happi@csd.uu.se): Created.
%% CVS:
%%    $Author: richardc $
%%    $Date: 2001/03/26 18:37:08 $
%%    $Revision: 1.1.1.1 $
%% ====================================================================
%% Exported functions (short description):
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(hipe_link_error).
-export([pp_error/1]).


pp_error({'EXIT',Err}) ->
  pp_exit(Err).

pp_exit({no_internal_function,Fun}) ->
  io:format("Could not find the address to the internal function ~w.\n",
	    [Fun]);
pp_exit({cant_link,no_hipe_link}) ->
  io:format("The bif hipe_link is missing, you must run ngorongoro(tm)\n");
pp_exit({immidiate_too_large,InstrList}) ->
  io:format("The following immidiates are to large:\n"),
  pp_imminstr_list(InstrList);
pp_exit({not_loaded,MFA}) ->
  case MFA of
    {M,F,A} ->
      io:format("The JAM-code of the function ~w:~w/~w is not loaded.\n"
		"Can't link.\n",
		[M,F,A]);
    Other -> io:format("~W doesn't seem to be a funktion.\n"
		       "Can't link.\n",[MFA])
  end;

pp_exit(ErrorList) when list(ErrorList) ->
 case lists:keysearch(problem,1,ErrorList) of 
   {value,{problem,Problem}} -> 
     case Problem of
       too_long_branch -> 
	 io:format("Branch of length ~w too long.",
		   [case lists:keysearch(length,1,ErrorList) of
		    {value,{length,Length}} -> Length;
		     _ -> unknown
		    end]),
	 pp_err_instr(ErrorList),
	 io:format(".\n",[]);
       load_address_should_be_removed ->
	 io:format("The instruction load_address is not hanled by the assebler.\n",
		   []);

       {memory_server,Error} ->
	 io:format("Something whent wrong when trying to allocate memory.\n",
		   []);
       {not_an_integer,Src} ->
	 io:format("~w in instruction ",[Src]),
	 pp_err_instr(ErrorList),
	 io:format("was supposted to be an integer.\n",[]);
       {not_handled,Unhandled} -> 
	 case Unhandled of
	   {address,Type} ->
	     io:format("The address is of an unhandled type in ",[]),
	     pp_err_instr(ErrorList),
	     io:format(".\n",[]);
	    {annulbit,b} ->
	     io:format("The handling of annulbits in b is not",[]),
	     io:format("implemented yet in ",[]),
	     pp_err_instr(ErrorList),
	     io:format(".\n",[]);		
	   {aluop,Op} ->
	     io:format("The ALU op ~w is not implemented.\nIn ",[Op]),
	     pp_err_instr(ErrorList),
	     io:format(".\n",[]);		
	   {aluccop,Op} ->
	     io:format("The ALUcc op ~w is not implemented.\nIn ",[Op]),
	     pp_err_instr(ErrorList),
	     io:format(".\n",[]);		
	   {cc,CC} ->
	     io:format("The condition code ~w is not implemented.\nIn ",[CC]),
	     pp_err_instr(ErrorList),
	     io:format(".\n",[]);
	   {br_cond,Cond} ->
	     io:format("The branch condition ~w is not implemented.\nIn ",[Cond]),
	     pp_err_instr(ErrorList),
	     io:format(".\n",[]);
	   {jmp_link,reg_reg_reg} ->
	     io:format("Jump link(resg,reg,reg) is not implemented.\nIn ",[]),
	     pp_err_instr(ErrorList),
	     io:format(".\n",[]);
	   {jmp,reg_reg} ->
	     io:format("Jump(reg,reg) is not implemented.\nIn ",[]),
	     pp_err_instr(ErrorList),
	     io:format(".\n",[]);
	   instruction ->
	     io:format("The instruction ",[]),
	     pp_err_instr(ErrorList),
	     io:format(" is not implemented.\n",[]);
	   Other ->
	     io:format("This ~w is not implemented, in ",[Other]),
	     pp_err_instr(ErrorList), 
	     io:format(".\n",[])
	 end;
       OtherProblem ->
	 io:format("The assembler exited with the problem ~w in instruction.",
		   [OtherProblem]),
	 
	 pp_err_instr(ErrorList), 
	 io:format(".\n",[])
     end;
   OtherError ->
     io:format("The assembler exited with the error ~w.",
	       [ErrorList]),
     io:format(".\n",[])
 end;
pp_exit(Error) ->
  io:format("The assembler exited with the error ~w.",
	    [Error]),
  io:format(".\n",[]).
	     
	   

pp_err_instr(ErrorList) ->
  case lists:keysearch(at,1,ErrorList) of 
    {value,{at,Instr}} ->
      hipe_sparc:pp_instr(Instr);
    
    _ -> io:format(" NoInstr ",[])
  end.
  

pp_imminstr_list([]) -> true;
pp_imminstr_list([{Instr,ImmList}|Rest]) ->
  io:format("In instruction",[]),
  hipe_sparc:pp_instr(Instr),
  io:format(":\n",[]),
  pp_imm_list(ImmList),
  pp_imminstr_list(Rest).

pp_imm_list([]) -> true;
pp_imm_list([Imm]) ->
  io:format(" ~w",Imm);
pp_imm_list([Imm|Rest]) ->
  io:format(" ~w,",Imm),
  pp_imm_list(Rest).

  

