%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Interface for register allocating SPARC code.  Uses regalloc.erl
%

-module(hipe_sparc_ls_or_gc_regalloc).
-export([alloc/2]).

-define(countspills,true).

-ifdef(countspills).
-define(count_spills(X), X).
-define(update_spillsum,
	case get(totalspill) of 
	  {__Loads,__Stores} ->
	    put(totalspill,{__Loads+get(loads),
			    __Stores+get(stores)}),
	    case get(spilledtemps) of
	      __Temps -> 
		put(spilledtemps, __Temps+get(temps));
	      _ ->
		true
	    end;
	  _ -> true 
	end).
-else.
-define(count_spills(X), true).
-define(update_spillsum, true).
-endif.

%
% Calls regalloc, rewrite the code after register allocation.
%
% Coloring are given as a list of {Reg, {reg, NewReg}} or 
% {Reg, {spill, SpillIndex}}.
%

alloc(CFG, Options) ->
  %% io:format("~w\n",[erlang:statistics(runtime)]),
  ?count_spills({put(loads,0),put(stores,0),put(temps,0)}),
  PhysRegs =   
    hipe_sparc_registers:allocatable() -- 
    hipe_sparc_registers:all_precolored(),
  Entrypoints = [hipe_sparc_cfg:start(SparcCfg)] ++ 
    hipe_sparc_cfg:fail_entrypoints(SparcCfg),

  Liveness = liveness(CFG),
  ?DEBUG(time,io:format("liveness (done)~w\n",[erlang:statistics(runtime)])),
  USIntervals = calculate_intervals(CFG,Liveness,Entrypoints, Options),
  ?DEBUG(time,io:format("intervals (done) ~w\n",[erlang:statistics(runtime)])),
  Intervals = sort_on_start(USIntervals),
  ?DEBUG(time,io:format("sort intervals (done) ~w\n",[erlang:statistics(runtime)])),
  ?DEBUG(time,io:format("No intervals: ~w\n",[length(Intervals)])),
  ?count_regs(put(noregs,get(noregs)+length(Intervals))),
  ?DEBUG(time,io:format("count intervals (done) ~w\n",[erlang:statistics(runtime)])),
  Allocation = allocate(Intervals,PhysRegs),
  ?DEBUG(time,io:format("allocation (done) ~w\n",[erlang:statistics(runtime)])),
  Allocation.

  {Coloring, NewSpillIndex} = 
    hipe_ls_regalloc:regalloc(SparcCfg,
				  hipe_sparc_registers:allocatable() -- 
				  (
				   hipe_sparc_registers:all_precolored()),
				  [hipe_sparc_cfg:start(SparcCfg)] ++ 
				  hipe_sparc_cfg:fail_entrypoints(SparcCfg),
				 Options),

  case NewSpillIndex > 0 of
    false ->
      ColTuple = hipe_sparc_ls_regalloc:cols2tuple(Coloring),
      Labels = hipe_sparc_cfg:labels(SparcCfg),
      hipe_sparc_ls_regalloc:rewrite(SparcCfg,ColTuple);
    true ->
      hipe_sparc_regalloc:alloc(SparcCfg)
  end.

