%%% $Id: hipe_x86_assemble.erl,v 1.60 2001/06/18 09:46:53 ulma4761 Exp $
%%% hipe_x86_assemble.erl
%%% Copyright (C) 2001 Ulf Magnusson
%%% Email: ulf.magnusson@ubm-computing.com
%%% This file given a pseudo x86 code will generate an erlang atom=(object format)
%%% containing the compiled code and patch referances

-module(hipe_x86_assemble).
-export([assemble/3]).
-include("../main/hipe.hrl").
-include("hipe_x86.hrl").
-include("../../kernel/src/hipe_ext_format.hrl").

% This function takes code and a Const table and creates an object file
% Input: CompiledCode, Closures, Flags
% Generate the data segment
% Generate the code segment
%  preprocess code,
%  calc size, remove labels, (remove comments) for each code seg.
%  Create labeltable (with offset information)
% Generate patchlist

assemble( CompiledCode, Closures, Flags ) ->
    put(hipe_x86_flags,Flags),
    print("****************** Assembling *******************\n"),
    % Want to have Code = [{MFA,Code,Consttab}|Rest] for compability
    NewCode = [{element(1,E), 
		{{hot,element(4,element(2,E))},[]}, 
		element(5,element(2,E))}
	       || E <- CompiledCode],

    {ConstSize,ConstMap, RefsFromConsts} = hipe_pack_constants:pack_constants( NewCode ),
    {Map,Code,PSJ} = get_code(NewCode),
    {SJmpTrue,HAddr,CAddr,AccHCode,AccCCode,AccHRefs,AccCRefs,FHN,FCN,NewMap} = mk_asmrefs(Code,Map,ConstMap,PSJ),
    print("Total num bytes=~w\n",[HAddr]),
    print("****************Assembler finished*****************\n"),
    AsmCode = {{HAddr,AccHCode,AccHRefs}, % SIZE is changed!!
    	       {CAddr,AccCCode,AccCRefs}}, % There are no cold code!
    %print("Constsize:~w\nConstMap:~w\n",[ConstSize, ConstMap]),
    Bin = mk_external(ConstSize, ConstMap, RefsFromConsts, NewMap, AsmCode,
		      Closures,SJmpTrue,FHN,FCN),
    %io:format("Assembler finished!\n"),
    Bin.
    
mk_external(ConstSize, ConstMap, RefsFromConsts, ExportMap, AsmCode, Closures,SJmpTrue,FHN,FCN) ->
  {{HotSize,AccHCode,AccHRefs},
   {ColdSize,AccCCode,AccCRefs}} = AsmCode,
  term_to_binary([?version(),
		  ConstSize,
		  slim_constmap(ConstMap),
		  mk_labelmap(RefsFromConsts, ExportMap, HotSize, SJmpTrue ),
		  slim_exportmap(ExportMap, HotSize, Closures, SJmpTrue,FHN,FCN),
		  HotSize,AccHCode,
		  slim_refs(AccHRefs),
		  ColdSize,AccCCode,
		  slim_refs(AccCRefs)]).

find_short_jumps(Map,[{MFA,Size,L,Block,Instr}|PSJs],AccSJmpTrue) ->
    {Seg,Address} = find( {MFA, L},Map),
    Offset = Address-Size-5, %BUG that don't trigger. FIXME
    case (Offset =< 127) and (Offset >= -128) of
	true ->
	    find_short_jumps(Map,PSJs,[{Size,Instr}|AccSJmpTrue]);
	false ->
	    find_short_jumps(Map,PSJs,AccSJmpTrue)
    end;
find_short_jumps(Map,[],AccSJmpTrue) ->
    AccSJmpTrue.

% all sjmptrue entries after AtPos must change Addr values to Addr-FewerBytes
modify_sjmptrue( SJmpTrue, 0, AtPos, NewSJmpTrue ) ->
    SJmpTrue;
modify_sjmptrue( [{Addr,Instr}|Rest], FewerBytes, AtPos, NewSJmpTrue ) ->
    case Addr >= AtPos of
	true ->
	    modify_sjmptrue( Rest, FewerBytes, AtPos, [{Addr-FewerBytes,Instr}|NewSJmpTrue] );
	false ->
	    modify_sjmptrue( Rest, FewerBytes, AtPos, [{Addr,Instr}|NewSJmpTrue] )
    end;
modify_sjmptrue( [], FewerBytes, AtPos, NewSJmpTrue ) ->
    NewSJmpTrue.

%Fix the labels
modify_labels( Map, 0, AtPos, NewMap ) ->
    Map;
modify_labels( [{MFA,Seg,Addr}|Rest], FewerBytes, AtPos, NewMap ) ->
    case Addr >= AtPos of
	true ->
	    modify_labels( Rest, FewerBytes, AtPos, [{MFA,Seg,Addr-FewerBytes}|NewMap] );
	false ->
	    modify_labels( Rest, FewerBytes, AtPos, [{MFA,Seg,Addr}|NewMap] )
    end;
modify_labels( [], FewerBytes, AtPos, NewMap ) ->
    NewMap.

call_align( [I|Is], Addr, AccCode, FewerNops, SJmpTrue, FakeAddr,Map ) ->
    Type = hipe_x86:insn_type(I),
    case Type of
	call ->
	    % Do call alignment
	    Size = accurat_size_of_instr(I,SJmpTrue,FakeAddr),
	    NumNopsNeeded = (4 - ((Size+Addr) rem 4)) rem 4,
	    ExtraNops = lists:duplicate( NumNopsNeeded, #nop{} ),

	    NewSJmpTrue = modify_sjmptrue(SJmpTrue,(3-NumNopsNeeded),FakeAddr+1,[] ),
	    NewMap = modify_labels( Map, (3-NumNopsNeeded), FakeAddr+1, [] ),

	    call_align( Is, Addr+accurat_size_of_instr(I,NewSJmpTrue,FakeAddr)+NumNopsNeeded, 
			[I|ExtraNops++AccCode], FewerNops+(3-NumNopsNeeded), NewSJmpTrue, 
			FakeAddr+instr_size(I)+NumNopsNeeded, NewMap );
	Other ->
	    call_align( Is, Addr+accurat_size_of_instr(I,SJmpTrue,FakeAddr), 
			[I|AccCode], FewerNops, SJmpTrue, FakeAddr+instr_size(I), Map )
    end; 
call_align( [], Addr, AccCode, FewerNops,SJmpTrue,FakeAddr, Map ) ->
    {lists:reverse(AccCode),FewerNops,SJmpTrue,Map}.

% return SJmpTrue so that the MAP can be corrected accordingly
% return AlignmentMap=FHN, one entry per mfa with how much alignment it should be corrected with
%                      this includes both mfa and call alignment.
mk_asmrefs(Code,Map,ConstMap,PSJ) ->
    SJmpTrue = find_short_jumps(Map,PSJ,[]),
    %print("True short jumps:~w\n",[SJmpTrue]),
    mk_asmrefs(Code,0,0,Map,ConstMap,[],[],[],[],SJmpTrue,0,0,[],0,0).

%only supports Hot Code!
mk_asmrefs([{MFA,Hot,Cold}|Rest],HAddr,CAddr,Map,ConstMap,AccHCode,AccCCode,
	   AccHRefs,AccCRefs,SJmpTrue,FakeHAddr,FakeCAddr, FHNops, HTotal,TotalHAlign ) ->
    print("Generating code for:~w\n",[MFA]),
    % Do aligning 4 byte.
    HAlign = (4 - (HAddr rem 4)) rem 4,
    ExtraHBytes = lists:duplicate( HAlign, 144 ), % insert nops in padding
    
    NewSJmpTrue = modify_sjmptrue(SJmpTrue,-HAlign,FakeHAddr,[] ),
    NewMap2 = modify_labels( Map,-HAlign, FakeHAddr, [] ),

    {NewHot,FewerHNops,NewHSJmpTrue,NewMap}   = call_align(Hot,HAddr+HAlign,[],0,NewSJmpTrue,FakeHAddr+HAlign,NewMap2 ),

    print( "Offset   | Opcode (hex)             | Instruction\n" ),
    {HCode,NewHAddr,HRefs,NewHFakeAddr} = mk_asmref(NewHot,MFA,HAddr+HAlign,NewMap,ConstMap,AccHRefs,
						    [],hot,NewHSJmpTrue,FakeHAddr+HAlign),
    NewCAddr = 0,
    CRefs = [],
    NewCFakeAddr = 0,

    print("Finished.\n\n"),
    mk_asmrefs(Rest,NewHAddr,NewCAddr,NewMap,ConstMap,AccHCode++ExtraHBytes++HCode,
	       [],HRefs,CRefs,NewHSJmpTrue,NewHFakeAddr,NewCFakeAddr,
	       [HAlign+TotalHAlign-HTotal|FHNops],HTotal+FewerHNops,TotalHAlign+HAlign);
mk_asmrefs([],HAddr,CAddr,Map,ConstMap,AccHCode,AccCCode,AccHRefs,AccCRefs,
	   SJmpTrue,HFAddr,CFAddr,FHN,HTotal,TotalHAlign) ->
  {SJmpTrue,HAddr,CAddr,AccHCode,AccCCode,AccHRefs,AccCRefs,FHN,FHN,Map}.

% This little function returns the size of a instruction
% Must check certain parameters for accurate answer.
% hmm, integrate with mk_asmref function possibly for avoiding double work.
% ask encode module of sizes.
% This function does not neccecerely return the final sizes, but more "fake" sizes
% since the short jump optimization is taken care after this.
instr_size( Instr ) ->
    Type = hipe_x86:insn_type(Instr),
    
    case Type of
	call -> 	    
	    {Arg,_} = jmp_resolve_arg( hipe_x86:call_fun(Instr), nomfa, 0, [], [],[],0, 0 ),
	    hipe_x86_encode:insn_sizeof(call,{Arg});
	jmp_fun -> 	    
	    {Arg,_} = jmp_resolve_arg( hipe_x86:jmp_fun_fun(Instr), nomfa, 0, [], [],[],0, 0 ),
	    hipe_x86_encode:insn_sizeof(jmp,{Arg});
	jmp_label -> 5; % always rel32 fake version!
	jmp_switch -> 7;
	cmp ->
	    {Arg,_} = resolve_alu_args(hipe_x86:cmp_src(Instr),hipe_x86:cmp_dst(Instr), 0, 0, [], nopatch,[]),
	    Size = hipe_x86_encode:insn_sizeof(cmp,Arg),
	    Size;
	move ->
	    {Arg,_} = resolve_move_args(hipe_x86:move_src(Instr),hipe_x86:move_dst(Instr), 0, 0, [], nopatch, nomfa, [] ),
	    hipe_x86_encode:insn_sizeof(mov,Arg);
	ret -> 3; % Always 3 bytes with imm16 value
	alu ->
	    {Arg,_} = resolve_alu_args(hipe_x86:alu_src(Instr),hipe_x86:alu_dst(Instr), 0, 0, [], nopatch,[]),
	    Size = hipe_x86_encode:insn_sizeof(hipe_x86:alu_op(Instr),Arg);
	jcc ->
	    case hipe_x86:jcc_label(Instr) of
		{x86_temp,Reg,_} ->
		    RM32 = {rm32,hipe_x86_encode:rm_reg( Reg )},
		    I = {jcc, {{cc,hipe_x86_encode:cc(hipe_x86:jcc_cond(Instr))},RM32 }},
		    hipe_x86_encode:insn_sizeof(I);
		{x86_mem,{x86_temp,TempDst,_},{x86_imm,ImmDst},_} ->
		    {EA,_} = create_ea(TempDst,ImmDst),
		    RM32 = {rm32,hipe_x86_encode:rm_mem(EA)},
		    I = {jcc, {{cc,hipe_x86_encode:cc(hipe_x86:jcc_cond(Instr))},RM32 }},
		    hipe_x86_encode:insn_sizeof(I);
		{x86_mfa,M,F,A} ->
		    6;
		{x86_prim,Prim} ->
		    6;
		_ ->
		    Arg = hipe_x86:jcc_label(Instr),
		    if
			integer(Arg) -> 6;
			atom(Arg) -> 6;
			true ->
			    ?EXIT(notsupportedjmpargument)
		    end
	    end;
        nop -> 1;
	push -> 
	    {Arg,_} = resolve_arg(hipe_x86:push_src(Instr),nomfa, 0, [], [], 0, [], 0, nopatch),
	    NewArg =
		case Arg of
		    {imm8,Imm} ->
			{imm32,Imm};
		    Other ->
			Other
		end,		    
	    Size = hipe_x86_encode:insn_sizeof( push,{NewArg} );
	inc ->
	    {Arg,_} = resolve_arg(hipe_x86:inc_dst(Instr),nomfa, 0, [], [], 0, [],0,nopatch),
	    Size = hipe_x86_encode:insn_sizeof( inc,{Arg} );
	dec ->
	    {Arg,_} = resolve_arg(hipe_x86:dec_dst(Instr),nomfa, 0, [], [], 0, [],0,nopatch),
	    Size = hipe_x86_encode:insn_sizeof( dec,{Arg} );
	cmovcc ->
	    {Arg,_} = resolve_move_args(hipe_x86:cmovcc_src(Instr),hipe_x86:cmovcc_dst(Instr),
					0, 0, [],nopatch, nomfa, [] ),
	    NewInstr = {cmovcc, {{cc,hipe_x86_encode:cc(hipe_x86:cmovcc_cond(Instr))}, Arg }},
	    Size = hipe_x86_encode:insn_sizeof(NewInstr);
	lea ->
	    {Arg,Ref} =	resolve_lea_args(hipe_x86:lea_mem(Instr),hipe_x86:lea_temp(Instr)),
	    Size = hipe_x86_encode:insn_sizeof(lea,Arg);
	label ->
	    0;
        Other -> ?EXIT({sizeOfInstructionNotSupported,Other})
    end.

% This function returns the true size of a instruction
accurat_size_of_instr( Instr,SJmpTrue,FakeAddr ) ->
    Type = hipe_x86:insn_type(Instr),
    
    case Type of
	call -> 	    
	    {Arg,_} = jmp_resolve_arg( hipe_x86:call_fun(Instr), nomfa, 0, [], [],[],0, 0 ),
	    hipe_x86_encode:insn_sizeof(call,{Arg});
	jmp_fun -> 	    
	    {Arg,_} = jmp_resolve_arg( hipe_x86:jmp_fun_fun(Instr), nomfa, 0, [], [],[],0, 0 ),
	    hipe_x86_encode:insn_sizeof(jmp,{Arg});
	jmp_label ->
	    case is_member(FakeAddr,SJmpTrue) of
		true ->
		    2;
		false ->
		    5
	    end;
	jmp_switch -> 7;
	cmp ->
	    {Arg,_} = resolve_alu_args(hipe_x86:cmp_src(Instr),hipe_x86:cmp_dst(Instr), 0, 0, [], nopatch,[]),
	    Size = hipe_x86_encode:insn_sizeof(cmp,Arg),
	    Size;
	move ->
	    {Arg,_} = resolve_move_args(hipe_x86:move_src(Instr),hipe_x86:move_dst(Instr), 0, 0, [], nopatch, nomfa, [] ),
	    hipe_x86_encode:insn_sizeof(mov,Arg);
	ret -> 3; % Always 3 bytes with imm16 value
	alu ->
	    {Arg,_} = resolve_alu_args(hipe_x86:alu_src(Instr),hipe_x86:alu_dst(Instr), 0, 0, [], nopatch,[]),
	    Size = hipe_x86_encode:insn_sizeof(hipe_x86:alu_op(Instr),Arg);
	jcc ->
	    case hipe_x86:jcc_label(Instr) of
		{x86_temp,Reg,_} ->
		    RM32 = {rm32,hipe_x86_encode:rm_reg( Reg )},
		    I = {jcc, {{cc,hipe_x86_encode:cc(hipe_x86:jcc_cond(Instr))},RM32 }},
		    hipe_x86_encode:insn_sizeof(I);
		{x86_mem,{x86_temp,TempDst,_},{x86_imm,ImmDst},_} ->
		    {EA,_} = create_ea(TempDst,ImmDst),
		    RM32 = {rm32,hipe_x86_encode:rm_mem(EA)},
		    I = {jcc, {{cc,hipe_x86_encode:cc(hipe_x86:jcc_cond(Instr))},RM32 }},
		    hipe_x86_encode:insn_sizeof(I);
		{x86_mfa,M,F,A} ->
		    6;
		{x86_prim,Prim} ->
		    6;
		_ ->
		    Arg = hipe_x86:jcc_label(Instr),
		    if
			integer(Arg) ->
			    	    case is_member(FakeAddr,SJmpTrue) of
					true -> 2;
					false -> 6
				    end;
			atom(Arg) -> 6;
			true ->
			    ?EXIT(notsupportedjmpargument)
		    end
	    end;
        nop -> 1; %Always one
	push -> 
	    {Arg,_} = resolve_arg(hipe_x86:push_src(Instr),nomfa, 0, [], [], 0, [], 0, nopatch),
	    NewArg =
		case Arg of
		    {imm8,Imm} ->
			{imm32,Imm};
		    Other ->
			Other
		end,		    
	    Size = hipe_x86_encode:insn_sizeof( push,{NewArg} );
	inc ->
	    {Arg,_} = resolve_arg(hipe_x86:inc_dst(Instr),nomfa, 0, [], [], 0, [],0,nopatch),
	    Size = hipe_x86_encode:insn_sizeof( inc,{Arg} );
	dec ->
	    {Arg,_} = resolve_arg(hipe_x86:dec_dst(Instr),nomfa, 0, [], [], 0, [],0,nopatch),
	    Size = hipe_x86_encode:insn_sizeof( dec,{Arg} );
	cmovcc ->
	    {Arg,_} = resolve_move_args(hipe_x86:cmovcc_src(Instr),hipe_x86:cmovcc_dst(Instr),
					0, 0, [],nopatch, nomfa, [] ),
	    NewInstr = {cmovcc, {{cc,hipe_x86_encode:cc(hipe_x86:cmovcc_cond(Instr))}, Arg }},
	    Size = hipe_x86_encode:insn_sizeof(NewInstr);
	lea ->
	    {Arg,Ref} =	resolve_lea_args(hipe_x86:lea_mem(Instr),hipe_x86:lea_temp(Instr)),
	    Size = hipe_x86_encode:insn_sizeof(lea,Arg);
	label ->
	    0;
        Other -> ?EXIT({sizeOfInstructionNotSupported,Other})
    end.
 
% This function looks at a single instruction,
% Assembles and possibly adds a patch to the patchlist
% which the loader will fix later.
% return: {assembled code, size, patches,fakesize}
mk_asmref([I|Is],MFA,Addr,Map,ConstMap,Refs,CodeList,Seg,SJmpTrue,FakeAddr) ->
    Type = hipe_x86:insn_type(I),
    print("~8s | ",[hipe_converters:int_to_hex(Addr)]),

    case Type of
	call ->
	    {Arg,Ref} = jmp_resolve_arg( hipe_x86:call_fun(I), MFA, Addr, Map, ConstMap,SJmpTrue,FakeAddr, 0 ),
	    Size = hipe_x86_encode:insn_sizeof(call,{Arg}),
	    Code = hipe_x86_encode:insn_encode(call,{Arg}),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Ref++Refs,
		      lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);
	jmp_fun ->
	    %sometimes via register
	    {Arg,Ref} = jmp_resolve_arg( hipe_x86:jmp_fun_fun(I), MFA, Addr, Map, ConstMap,SJmpTrue,FakeAddr, 0 ),
	    Code = hipe_x86_encode:insn_encode(jmp,{Arg}),
	    Size = hipe_x86_encode:insn_sizeof(jmp,{Arg}),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Ref++Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);
	jmp_label ->
	    %io:format("mk_asmref jmp_label Fakeaddr:~w\n",[FakeAddr]),
	    {Arg,Ref} = jmp_resolve_arg( hipe_x86:jmp_label_label(I), MFA, Addr, Map, ConstMap,SJmpTrue,FakeAddr, 0 ),
	    Code = hipe_x86_encode:insn_encode(jmp,{Arg}),
	    Size = hipe_x86_encode:insn_sizeof(jmp,{Arg}),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Ref++Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+instr_size(I));
	jmp_switch ->
	    %First the encoding of the instruction..
	    SINDEX = hipe_x86_encode:sindex( 2, element(2,hipe_x86:jmp_switch_temp(I)) ),
	    EA = hipe_x86_encode:ea_disp32_sindex( 0 ,SINDEX ), % this creates a SIB implicitly
	    RM32 = {rm32,hipe_x86_encode:rm_mem(EA)},
	    Code = hipe_x86_encode:insn_encode(jmp,{RM32}),
	    Size = hipe_x86_encode:insn_sizeof(jmp,{RM32}),
	    
	    %Then the patching...
	    %io:format("Constant Map:~w\n",[ConstMap]),
	    ConstNo = find_const({MFA,hipe_x86:jmp_switch_jtab(I)},ConstMap),
	    Ref = {?PATCH_TYPE2EXT(load_address),Addr+3,{constant,ConstNo}},
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,[Ref|Refs],lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);
	cmp ->
	    {Arg,Ref} = resolve_alu_args(hipe_x86:cmp_src(I),hipe_x86:cmp_dst(I), Addr, FakeAddr, SJmpTrue, patch,ConstMap ),
	    Code = hipe_x86_encode:insn_encode(cmp,Arg),
	    Size = hipe_x86_encode:insn_sizeof(cmp,Arg),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Ref++Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);
	move ->
	    {Arg,Ref} =
		resolve_move_args(hipe_x86:move_src(I),hipe_x86:move_dst(I),
			    Addr, FakeAddr, SJmpTrue,patch, MFA, ConstMap ),
	    Code = hipe_x86_encode:insn_encode(mov,Arg),
	    Size = hipe_x86_encode:insn_sizeof(mov,Arg),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Ref++Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);
	ret -> 
	    Code = hipe_x86_encode:insn_encode( ret,{{imm16,hipe_x86:ret_npop(I)}} ),
	    Size = hipe_x86_encode:insn_sizeof( ret,{{imm16,hipe_x86:ret_npop(I)}} ),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);
	% add,sub,cmp,...
	alu ->
	    {Arg,Ref} = resolve_alu_args( hipe_x86:alu_src(I),hipe_x86:alu_dst(I), Addr, FakeAddr, SJmpTrue, patch, ConstMap ),
	    Code = hipe_x86_encode:insn_encode(hipe_x86:alu_op(I),Arg),
	    Size = hipe_x86_encode:insn_sizeof(hipe_x86:alu_op(I),Arg),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Ref++Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);
	% Conditinal jumps param = {{cc,CC}, {rel32,Rel32}}
	jcc ->
	    {Arg,Ref} = jmp_resolve_arg( hipe_x86:jcc_label(I), MFA, Addr, Map, ConstMap,SJmpTrue,FakeAddr, 1),
	    Instr = {jcc, {{cc,hipe_x86_encode:cc(hipe_x86:jcc_cond(I))}, Arg }},
	    Code = hipe_x86_encode:insn_encode( Instr ),
	    Size = hipe_x86_encode:insn_sizeof( Instr ),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Ref++Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+instr_size(I));
	nop ->
	    Code = hipe_x86_encode:insn_encode( nop,{} ),
	    Size = hipe_x86_encode:insn_sizeof( nop,{} ),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);
	push ->
	    %io:format("push arg:~w\n",[hipe_x86:push_src(I)]),
	    {Arg,Ref} = resolve_arg(hipe_x86:push_src(I),MFA, Addr, Map, ConstMap, FakeAddr, SJmpTrue,1,patch),
	    % Make Sure we push 32 bit
	    NewArg =
		case Arg of
		    {imm8,Imm} ->
			{imm32,Imm};
		    Other ->
			Other
		end,		    
	    Code = hipe_x86_encode:insn_encode( push,{NewArg} ),
	    Size = hipe_x86_encode:insn_sizeof( push,{NewArg} ),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Ref++Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);
	inc ->
	    % increase operation
	    {Arg,Ref} = resolve_arg(hipe_x86:inc_dst(I),MFA, Addr, Map, ConstMap, FakeAddr, SJmpTrue,1,patch),
	    Code = hipe_x86_encode:insn_encode( inc,{Arg} ),
	    Size = hipe_x86_encode:insn_sizeof( inc,{Arg} ),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Ref++Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);
	dec ->
	    % decrease
	    {Arg,Ref} = resolve_arg(hipe_x86:dec_dst(I),MFA, Addr, Map, ConstMap, FakeAddr, SJmpTrue,1,patch),
	    Code = hipe_x86_encode:insn_encode( dec,{Arg} ),
	    Size = hipe_x86_encode:insn_sizeof( dec,{Arg} ),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Ref++Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);
	cmovcc ->
	    % Conditional move
	    {Arg,Ref} =	resolve_move_args(hipe_x86:cmovcc_src(I),hipe_x86:cmovcc_dst(I),
					  Addr, FakeAddr, SJmpTrue,patch, MFA, ConstMap ),
	    Instr = {cmovcc, {{cc,hipe_x86_encode:cc(hipe_x86:cmovcc_cond(I))}, Arg }},
	    Code = hipe_x86_encode:insn_encode(Instr),
	    Size = hipe_x86_encode:insn_sizeof(Instr),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Ref++Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);
	lea ->
	    {Arg,Ref} =	resolve_lea_args(hipe_x86:lea_mem(I),hipe_x86:lea_temp(I)),
	    Code = hipe_x86_encode:insn_encode(lea,Arg),
	    Size = hipe_x86_encode:insn_sizeof(lea,Arg),
	    print_code_list(Code),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr+Size,Map,ConstMap,Ref++Refs,lists:reverse(Code,CodeList),Seg,SJmpTrue,FakeAddr+Size);	    
	label ->
	    print_code_list([]),
	    print_insn(I),
	    mk_asmref(Is,MFA,Addr,Map,ConstMap,Refs,CodeList,Seg,SJmpTrue,FakeAddr)
    end;

mk_asmref([],_,Addr,_,_,Refs,Code,_,_,FakeAddr) ->
    %print("mk_asmref finished, fixing list\n" ),
    {lists:reverse(Code),Addr,Refs,FakeAddr}.

create_ea(Base, Off) ->
    if 
	Off =:= 0 ->
	    case Base of
		4 -> %esp, use SIB w/o disp8
		    SIB = hipe_x86_encode:sib(Base),
		    {hipe_x86_encode:ea_sib(SIB), -3};
		5 -> %ebp, use disp8 w/o SIB
		    {hipe_x86_encode:ea_disp8_base(Off, Base), -3};
		_ -> %neither SIB nor disp8 needed
		    {hipe_x86_encode:ea_base(Base), -4}
	    end;
       Off >= -128, Off =< 127 ->
	    case Base of
		4 -> %esp, must use SIB
		    SIB = hipe_x86_encode:sib(Base),
		    {hipe_x86_encode:ea_disp8_sib(Off, SIB), -2};
		_ -> %use disp8 w/o SIB
		    {hipe_x86_encode:ea_disp8_base(Off, Base), -3}
	    end;
       true ->
	    case Base of
		4 -> %esp, must use SIB
		    SIB = hipe_x86_encode:sib(Base),
		    {hipe_x86_encode:ea_disp32_sib(Off, SIB), 1};
		_ ->
		    {hipe_x86_encode:ea_disp32_base(Off, Base), 0}
	    end
    end.

% lea reg, mem
resolve_lea_args( {x86_mem,{x86_temp,TempSrc,_},{x86_imm,ImmSrc},_}, {x86_temp,Temp,_} ) ->
    {EA,Extra} = create_ea(TempSrc,ImmSrc),
    {{{reg32,Temp},{ea,EA}},[]}.

% mov mem, imm
resolve_move_args( {x86_imm,ImmSrc}, {x86_mem,{x86_temp,Temp,_},{x86_imm,ImmDst},_}, 
		   Addr, FakeAddr, SJmpTrue,Patch,MFA, ConstMap ) ->
    {EA,Extra} = create_ea(Temp,ImmDst),
    RM32 = {rm32,hipe_x86_encode:rm_mem(EA)},
    {{_,Imm},Ref} = resolve_arg( {x86_imm,ImmSrc}, MFA, Addr, [], ConstMap, FakeAddr, SJmpTrue,6+Extra,Patch ),
    {{RM32,{imm32,Imm}},Ref};
% mov reg,mem
resolve_move_args( {x86_mem,{x86_temp,TempSrc,_},{x86_imm,ImmSrc},_}, {x86_temp,Temp,_}, 
		   Addr, FakeAddr, SJmpTrue,Patch,MFA, ConstMap ) ->
    {EA,Extra} = create_ea(TempSrc,ImmSrc),
    RM32 = {rm32,hipe_x86_encode:rm_mem(EA)},
    {{{reg32,Temp},RM32},[]};
% mov mem,reg
resolve_move_args( {x86_temp,Temp,_}, {x86_mem,{x86_temp,TempDst,_},{x86_imm,ImmDst},_}, 
		   Addr, FakeAddr, SJmpTrue,Patch,MFA, ConstMap ) ->
    {EA,Extra} = create_ea(TempDst,ImmDst),
    RM32 = {rm32,hipe_x86_encode:rm_mem(EA)},
    {{RM32,{reg32,Temp}},[]};
% mov reg,reg
resolve_move_args( {x86_temp,SrcTemp,_}, {x86_temp,DstTemp,_}, 
		   Addr, FakeAddr, SJmpTrue,Patch,MFA, ConstMap ) ->
    RM32 = {rm32,hipe_x86_encode:rm_reg( SrcTemp )},
    {{{reg32,DstTemp},RM32},[]};
% mov reg,imm
resolve_move_args( Src={x86_imm,ImmSrc}, Dst={x86_temp,Temp,_}, Addr, FakeAddr, SJmpTrue,Patch,MFA, ConstMap ) ->
    {{_,Imm},Ref} = resolve_arg( Src, MFA, Addr, [], ConstMap, FakeAddr, SJmpTrue,1,Patch ),
    {{{reg32,Temp},{imm32,Imm}},[]}.

% ALU, takes basically the same variants as the move,
% however due to fewer encoding possibilities this must be a own function.
% alu mem, imm
resolve_alu_args( {x86_imm,ImmSrc}, {x86_mem,{x86_temp,Temp,_},{x86_imm,ImmDst},_}, Addr, FakeAddr, SJmpTrue,Patch, ConstMap ) ->
    {EA,Extra} = create_ea(Temp,ImmDst),
    RM32 = {rm32,hipe_x86_encode:rm_mem(EA)},
    {{_,Imm},Ref} = resolve_arg( {x86_imm,ImmSrc}, nomfa, Addr, [], ConstMap, FakeAddr, SJmpTrue,6+Extra,Patch ),
    case (Imm =< 127) and (Imm >= -128) of
	true ->
	    {{RM32,{imm8,Imm}},Ref};
	false ->
	    {{RM32,{imm32,Imm}},Ref}
    end;

% alu reg,mem
resolve_alu_args( {x86_mem,{x86_temp,TempSrc,_},{x86_imm,ImmSrc},_}, {x86_temp,Temp,_}, Addr, FakeAddr, SJmpTrue,Patch, ConstMap ) ->
    {EA,Extra} = create_ea(TempSrc,ImmSrc),
    RM32 = {rm32,hipe_x86_encode:rm_mem(EA)},
    {{{reg32,Temp},RM32},[]};
% alu mem,reg
resolve_alu_args( {x86_temp,Temp,_}, {x86_mem,{x86_temp,TempDst,_},{x86_imm,ImmDst},_}, Addr, FakeAddr, SJmpTrue,Patch, ConstMap ) ->
    {EA,Extra} = create_ea(TempDst,ImmDst),
    RM32 = {rm32,hipe_x86_encode:rm_mem(EA)},
    {{RM32,{reg32,Temp}},[]};
% alu reg,reg
resolve_alu_args( {x86_temp,SrcTemp,_}, {x86_temp,DstTemp,_}, Addr, FakeAddr, SJmpTrue,Patch, ConstMap  ) ->
    RM32 = {rm32,hipe_x86_encode:rm_reg( SrcTemp )},
    {{{reg32,DstTemp},RM32},[]};
% alu reg,imm
resolve_alu_args( Src={x86_imm,ImmSrc}, Dst={x86_temp,Temp,_}, Addr, FakeAddr, SJmpTrue,Patch, ConstMap ) ->
    {{_,Imm},Ref} = resolve_arg( Src, nomfa, Addr, [], ConstMap, FakeAddr, SJmpTrue, 2,Patch ),
    RM32 = {rm32,hipe_x86_encode:rm_reg(Temp)},
    {{RM32,{imm32,Imm}},[]}.

traverse_tagged_short_jump_list([{Addr,Instr}|Rest],Acc) ->
    case Instr of
	jcc ->
	    traverse_tagged_short_jump_list(Rest,Acc+4);
	jmp ->
	    traverse_tagged_short_jump_list(Rest,Acc+3)
    end;
traverse_tagged_short_jump_list([],Acc) ->
    Acc.

correct_dist_to_label( Start, End, List ) ->
    if
	Start<End ->
	    L = [E || {X,Instr} = E <- List,
		      (X >= Start) and (X < End)],
	    -traverse_tagged_short_jump_list(L,0);
	Start>End ->
	    L = [E || {X,Instr} = E <- List,
		      (X >= End) and (X < Start)],
	    traverse_tagged_short_jump_list(L,0);
	true ->
	    0
    end.

is_member(Addr, Instrs) -> lists:keymember(Addr, 1, Instrs).

% When patching is needed be sure to correct the address with short jumps decreases
% The Extra parameter is for jcc jumps which is 6 bytes not 5!
jmp_resolve_arg( Jmp_arg, MFA, Addr, Map, ConstMap, SJmpTrue,FakeAddr, Extra ) ->
    %io:format("Jmp_arg:~w\n",[Jmp_arg]),
    case Jmp_arg of
	{x86_temp,Reg,_} ->
	    %io:format("Jumping via register\n"),
	    RM32 = {rm32,hipe_x86_encode:rm_reg( Reg )},
	    {RM32,[]};
	{x86_mem,{x86_temp,TempDst,_},{x86_imm,ImmDst},_} ->
	    %io:format("Jumping via memory\n"),
	    {EA,_} = create_ea(TempDst,ImmDst),
	    RM32 = {rm32,hipe_x86_encode:rm_mem(EA)},
	    {RM32,[]};
	{x86_mfa,M,F,A} ->
	    {{rel32,0},[{?PATCH_TYPE2EXT(mfa),Addr+1,{M,F,A}}]};
	{x86_prim,Prim} ->
	    {{rel32,0},[{?PATCH_TYPE2EXT(call_bif),Addr+1,Prim}]};
	_ ->
	    if
	        % This is a label, find it, calculate relative offset and return it.
		integer(Jmp_arg) ->
		    {Seg,Address} = find( {MFA, Jmp_arg},Map),
		    case is_member(FakeAddr,SJmpTrue) of
			true ->
			    Offset = Address-FakeAddr-2 + correct_dist_to_label(FakeAddr,Address,SJmpTrue),
			    {{rel8,Offset},[]};
			false ->
			    Offset = Address-FakeAddr-5-Extra + correct_dist_to_label(FakeAddr,Address,SJmpTrue),
			    %io:format("Long jump, Address:~w FakeAddr:~w Correct:~w SjumpTrue:~w Offset:~w\n",
				%      [Address,FakeAddr,correct_dist_to_label(FakeAddr,Address,SJmpTrue),SJmpTrue,Offset]),
			    {{rel32,Offset},[]}
		    end;
		atom(Jmp_arg) ->
		    %io:format("Map:~w\n",[Map]),
		    % This is a bif, add reference to patchlist
		    {{rel32,0},[{?PATCH_TYPE2EXT(call_bif),Addr+1,Jmp_arg}]};
		true ->
		    %This should not happen
		    ?EXIT(notsupportedjmpargument)
	    end
    end.

% return {arg for encoding,Refs}
% BytesToImm32 is the offset from the first byte of the imm32 word.
% e.g call imm32, BytesToImm32=1
% Addr = Offset from byte zero in this module to the imm32 value.
resolve_arg( Arg, MFA, Addr, Map, ConstMap, FakeAddr, SJmpTrue, BytesToImm32,Patch ) ->
    case Arg of
	{reg32,Reg32} ->
	    % %if it is a reg no patching is needed
	    {Arg,[]};
        {x86_imm,Imm} ->
	    if
		atom(Imm) ->
		    %print("Atom:~w added to patchlist at addr:~w - ",[Imm,Addr]),
		    {{imm32,9},[{?PATCH_TYPE2EXT(load_atom),Addr+BytesToImm32,Imm}]};
		integer(Imm) ->
		    case (Imm < 128) and (Imm > -127) of % XXX: -128 <= Imm <= 127 ???
			true ->
			    {{imm8,Imm},[]};
			false ->
			    {{imm32,Imm},[]}
		    end;
		true ->
		    % Ok, this is the case when {label,type} (type=label,catch,constant)
		    case Patch of
			nopatch ->
			    {{imm32,0},[]};
			patch ->
			    case element(2,Imm) of
				% label wrong FIXME
				label -> % type
				    ?EXIT(notsupportinglabelcatch);
% 				    {Seg,Address} = find( {MFA, element(1,Imm)},Map),
% 				    Offset =
% 					case FakeAddr<Address of
% 					    true -> %jump forward use -2
% 						Address-Addr-BytesToImm32-4 + correct_dist_to_label(FakeAddr,Address,SJmpTrue);
% 					    false -> %jump backward
% 						%print("Address:~w Addr:~w FakeAddr:~w\n",[Address,Addr,FakeAddr]),
% 						Address-FakeAddr-BytesToImm32-4 - correct_dist_to_label(FakeAddr,Address,SJmpTrue)
% 					end,
% 				    print("label:~w added to patchlist at offset:~w\n",[element(1,Imm),Offset]),
				    
% 				    {{imm32,0},{?PATCH_TYPE2EXT(load_address),Addr+BytesToImm32,{label,Offset}}};
				% Constant
				constant -> % type
				    ConstNo = find_const({MFA,element(1,Imm)},ConstMap),
				    {{imm32,0},[{?PATCH_TYPE2EXT(load_address),Addr+BytesToImm32,{constant,ConstNo}}]};
				'catch' ->
				    % This is a catch
				    {Seg,Address} = find( {MFA, element(1,Imm)},Map),
				    Offset = Address + correct_dist_to_label(0,Address,SJmpTrue),
				    %print("Catch:~w label at offset:~w\n",[element(1,Imm),Offset]),
				    {{imm32,0},[{?PATCH_TYPE2EXT(load_address),Addr+BytesToImm32,{'catch',Offset}}]};
				closure ->
				    {{imm32,0},[{?PATCH_TYPE2EXT(load_address),Addr+BytesToImm32,
						{closure,element(1,Imm)}}]};
				Other ->
				    ?EXIT({unsupportedimm32value,Other})
			    end
		    end
	    end;
        {x86_temp,Temp,_} ->
	    {{reg32,Temp},[]};
	% Push uses this, and goes via ESP so the SIB byte stays...
	{x86_mem,{x86_temp,Temp,_},{x86_imm,Imm32},_} ->
	    {EA,Extra} = create_ea( Temp, Imm32 ),
	    DST = {rm32,hipe_x86_encode:rm_mem(EA)},
	    {DST,[]}
    end.

get_code(Code) ->
  get_code(Code,0,0,[],[],[]).

get_code([{MFA,Code,_}|Rest],HotAddress,ColdAddress,Map,AccCode,PSJ) ->
    {NewHot,NewCold,HotSize,ColdSize,NewMap,NewPSJ} = preprocess(Code,MFA,HotAddress,ColdAddress,Map,PSJ),
    get_code(Rest,HotSize,ColdSize,NewMap,[{MFA,NewHot,NewCold}|AccCode],NewPSJ);
get_code([],HotSize,ColdSize,Map,Acc,PSJ) -> 
    %%{HotSize,ColdSize,Map,lists:reverse(Acc)}.
    {Map,lists:reverse(Acc),PSJ}.
  
preprocess({{Block,Entry},Rest},MFA,HotAddress,ColdAddress,Map,PSJ) ->
%%  print("~w at ~w ~w\n",[MFA,HotAddress,ColdAddress]),
  process([{Block,Entry}|Rest],[],[],HotAddress,ColdAddress,
	      [{{MFA,entry},Block,
		case Block of
		  hot ->
		    HotAddress;
		 cold ->
		   ColdAddress
		end}|Map],
	  MFA,PSJ).


process([{Block,Code}|Rest],Hot,Cold,HotSize,ColdSize,Map,MFA,PSJ) ->
    case Block of
	hot ->
	    {NewCode,NewSize,NewMap,NewPSJ} = process_instr(Code,HotSize,Map,MFA,[],hot,PSJ),
	    %print("Labels found in code:\n~w",[NewMap]),
	    process(Rest,NewCode++Hot,Cold,NewSize,ColdSize,NewMap,MFA,NewPSJ);
	cold ->
	    {NewCode,NewSize,NewMap,NewPSJ} = process_instr(Code,ColdSize,Map,MFA,[],cold,PSJ),
	    process(Rest,Hot,NewCode++Cold,HotSize,NewSize,NewMap,MFA,NewPSJ)
    end;

process([],Hot,Cold,HotSize,ColdSize,Map,_,PSJ) ->
    {Hot,Cold,HotSize,ColdSize,Map,PSJ}.

process_instr([I|Is],Size,Map,MFA,AccCode,Block,PossibleShortJumps) ->
    %io:format("Addr:~s ~w\n",[hipe_converters:int_to_hex(Size),I]),
    %io:format("I=~w\n",[I]),
  case hipe_x86:insn_type(I) of
      jmp_label ->
	  L = hipe_x86:jmp_label_label(I),
	  if
	      integer(L) ->
		  % Possible two byte
		  process_instr(Is,Size+instr_size(I),Map,
				MFA,[I|AccCode],Block,[{MFA,Size,L,Block,jmp}|PossibleShortJumps]);
	      true ->
		  process_instr(Is,Size+instr_size(I),Map,
				MFA,[I|AccCode],Block,PossibleShortJumps)
	  end;
      jcc ->
	  L = hipe_x86:jcc_label(I),
	  if
	      integer(L) ->
		  % Possible two byte
		  process_instr(Is,Size+instr_size(I),Map,
				MFA,[I|AccCode],Block,[{MFA,Size,L,Block,jcc}|PossibleShortJumps]);
	      true ->
		  process_instr(Is,Size+instr_size(I),Map,
				MFA,[I|AccCode],Block,PossibleShortJumps)
	  end;
      label ->
	  %print("Found label:~w, at pos:~w adding to Map\n",[hipe_x86:label_label(I),Size]),
	  process_instr(Is,Size,[{{MFA,hipe_x86:label_label(I)},Block,Size}|Map],
			MFA,[I|AccCode],Block,PossibleShortJumps);
      call ->
	  % reserve space for possible three nops
	  process_instr(Is,Size+instr_size(I)+3,Map,MFA,[I|AccCode],Block,PossibleShortJumps);
      comment ->
	  process_instr(Is,Size,Map,MFA,AccCode,Block,PossibleShortJumps);
      Other ->
	  %print("PR Other instr:~w",[Other]),
	  %process_instr(Is,Size+hipe_x86_assemble:size_of_instr(I),Map,MFA,[I|AccCode],Block)
	  process_instr(Is,Size+instr_size(I),Map,MFA,[I|AccCode],Block,PossibleShortJumps)
  end;
process_instr([],Size,Map,_,Code,_,PSJ) ->
  {lists:reverse(Code),Size,Map,PSJ}.

slim_constmap(Map) ->
  slim_constmap(Map,[]).
slim_constmap([{MFA, Label, ConstNo,
		Offset, Need, Type, Term, Export}|Rest],Acc) ->
  slim_constmap(Rest, [ConstNo, Offset, Need, Type, Export, Term| Acc]);
slim_constmap([],Acc) -> Acc.

mk_labelmap(Map, ExportMap, HotSize, SJmpTrue ) ->
  %% msg("Map: ~w Map\n",[Map]),
  LblMap = lists:flatten(mk_labelmap(Map, ExportMap,[],HotSize, SJmpTrue )),
  %% msg("LblMap: ~w Map\n",[LblMap]),
  LblMap.

mk_labelmap([{MFA, Labels}| Rest], ExportMap, Acc, HotSize, SJmpTrue ) ->
    Map = [case Label of
	       {L,Pos} ->
		   Offset = find_offset({MFA,L}, ExportMap,HotSize),
		   %%io:format("1 SJmpTrue:~w\nOffset:~w\n",[SJmpTrue,Offset]),
		   Dist = correct_dist_to_label(0, Offset, SJmpTrue),
		   {Pos,Offset+Dist};
	       {sorted,Base,OrderedLabels} ->
		   {sorted, Base, [begin
				       Offset = find_offset({MFA,L},
							    ExportMap,
							    HotSize),
				       %%io:format("2 SJmpTrue:~w\nOffset:~w\n",[SJmpTrue,Offset]),
				       Dist = correct_dist_to_label(0,
								    Offset,
								    SJmpTrue),
				       {Order, Offset+Dist}
				   end
				   || {L,Order} <- OrderedLabels]}
	   end
	   || Label <- Labels],
    %% msg("Map: ~w Map\n",[Map]),
    mk_labelmap(Rest, ExportMap, [Map,Acc], HotSize, SJmpTrue);

mk_labelmap([],_,Acc,_, SJmpTrue ) -> Acc.

slim_exportmap(Map, HotSize, Closures, SJmpTrue, FHN, FCN ) ->
  SortedMap = lists:sort(slim_exportmap1(Map, HotSize, [], SJmpTrue, FHN,FCN)),
  
  slim_sorted_exportmap(SortedMap, Closures).
    

slim_exportmap1([{{{M,F,A},entry},Mode,Adr}|Rest], HotSize, Acc, SJmpTrue,[FHN|FHNs],[FCN|FCNs]) ->
  NewAdr =
    case Mode of
      hot -> Adr + correct_dist_to_label( 0, Adr, SJmpTrue );
      cold -> Adr + HotSize
    end,
    %io:format("MFA:~w NewAdr:~w Adr:~w SJmpTrue:~w\n",[{M,F,A},NewAdr,Adr,SJmpTrue]),
  slim_exportmap1(Rest, HotSize, [{NewAdr,M,F,A}|Acc], SJmpTrue,FHNs,FCNs);
slim_exportmap1([_|Rest], HotSize, Acc, SJmpTrue,FHN,FCN) ->
  slim_exportmap1(Rest, HotSize, Acc, SJmpTrue,FHN,FCN);
slim_exportmap1([], HotSize, Acc, SJmpTrue,[],[]) -> Acc.

slim_sorted_exportmap([{Adr,M,F,A}|Rest], Closures) ->
  IsClosure = lists:member({M,F,A}, Closures),
  [Adr, M,F,A, IsClosure | slim_sorted_exportmap(Rest, Closures)];
slim_sorted_exportmap([],_) -> [].



slim_refs([]) -> [];
slim_refs(Refs) ->
  [Ref|Rest] = lists:keysort(1,Refs),
  compact_ref_types(Rest,element(1,Ref),[Ref],[]).

compact_ref_types([Ref|Refs],Type,AccofType,Acc) ->
  case element(1,Ref) of
    Type ->
      compact_ref_types(Refs,Type,[Ref|AccofType], Acc);
    NewType ->
      compact_ref_types(Refs, NewType,[Ref], [{Type,compact_dests(AccofType)}|Acc])
  end;
compact_ref_types([],Type,AccofType,Acc) ->
  [{Type,compact_dests(AccofType)}|Acc].


compact_dests([]) -> [];
compact_dests(Refs) ->
  [Ref|Rest] = lists:keysort(3,Refs),
  compact_dests(Rest,element(3,Ref),[element(2,Ref)],[]).


compact_dests([Ref|Refs],Dest,AccofDest,Acc) ->
  case element(3,Ref) of
    Dest ->
      compact_dests(Refs,Dest,[element(2,Ref)|AccofDest], Acc);
    NewDest ->
      compact_dests(Refs, NewDest,[element(2,Ref)], [{Dest,AccofDest}|Acc])
  end;
compact_dests([],Dest,AccofDest,Acc) ->
  [{Dest,AccofDest}|Acc].

find(Name,[{Name,Seg,Addr}|_]) ->
  {Seg,Addr};
find(Name,[_|Rest]) ->
  find(Name,Rest);
find({Name,entry},[]) ->
  {none, 0}.

find_offset({MFA,L},[{{MFA,L},Mode,Adr}|Rest], HotSize) ->
  case Mode of
    hot -> Adr;
    cold -> Adr + HotSize
  end;
find_offset(L,[_|Rest], HotSize) ->
  find_offset(L,Rest,HotSize);
find_offset(L,[],_) ->
  exit({label_not_found,L}).

print( String ) ->
    Flags = get(hipe_x86_flags),
    ?when_option(assembly_listing, Flags,io:format(String,[])).
print( String, Arglist ) ->
    Flags = get(hipe_x86_flags),
    ?when_option(assembly_listing, Flags,io:format(String,Arglist)).

print_insn(I) ->
    Flags = get(hipe_x86_flags),
    ?when_option(assembly_listing, Flags,hipe_x86_pp:pp_insn(I)).
   
fill_spaces( 0 ) ->
    ok;
fill_spaces( Num ) ->
    io:format(" "),
    fill_spaces( Num-1 ).

print_code_list(Code) ->
    Flags = get(hipe_x86_flags),
    ?when_option(assembly_listing, Flags, 
		 lists:foreach( 
		   fun(B) -> if B<0 ->
				     io:format("~s",[hipe_converters:int_to_hex(256+B)]);
				true ->
				     if B<16 ->
					     io:format("0~s",[hipe_converters:int_to_hex(B)]);
					true ->
					     io:format("~s",[hipe_converters:int_to_hex(B)])
				     end
			     end
		   end,
		   Code)),
    % Print the rest 12-size(code) with spaces. and a space and a | and a space
    ?when_option(assembly_listing, Flags,fill_spaces( 24-(length(Code)*2) ) ),
    ?when_option(assembly_listing, Flags,io:format(" | ") ).

find_const({MFA,Label},[{MFA,Label,ConstNo,_,_, _, _,_}|_])->
  ConstNo;
find_const(N,[M|R]) ->
  find_const(N,R);
find_const(C,[]) ->
  ?EXIT({constant_not_found,C}).

