%%% $Id: hipe_x86_defuse.erl,v 1.7 2001/06/13 12:43:33 mikpe Exp $
%%% compute def/use sets for x86 insns
%%%
%%% TODO:
%%% - represent EFLAGS (condition codes) use/def by a virtual reg?
%%% - should call def the clobbered regs?
%%% - should push use/def %esp?

-module(hipe_x86_defuse).
-export([insn_def/1, insn_use/1, src_use/1]).
-include("hipe_x86.hrl").

%%%
%%% insn_def(Insn) -- Return set of temps defined by an instruction.
%%%

insn_def(I) ->
    case I of
	#alu{dst=Dst} -> dst_def(Dst);
	#cmovcc{dst=Dst} -> dst_def(Dst);
	#dec{dst=Dst} -> dst_def(Dst);
	#inc{dst=Dst} -> dst_def(Dst);
	#lea{temp=Temp} -> [Temp];
	#move{dst=Dst} -> dst_def(Dst);
	#pseudo_call{dsts=Dsts} -> addtemps(Dsts, []);
	%% call, cmp, comment, jcc, jmp_fun, jmp_label, jmp_switch, label
	%% nop, pseudo_jcc, pseudo_return, pseudo_tailcall, push, ret
	_ -> []
    end.

dst_def(Dst) ->
    case Dst of
	#x86_temp{} -> [Dst];
	_ -> []
    end.

%%%
%%% insn_use(Insn) -- Return set of temps used by an instruction.
%%%

insn_use(I) ->
    case I of
	#alu{src=Src,dst=Dst} -> addtemp(Src, addtemp(Dst, []));
	#call{'fun'=Fun} -> addtemp(Fun, []);
	#cmovcc{src=Src, dst=Dst} -> addtemp(Src, dst_use(Dst));
	#cmp{src=Src, dst=Dst} -> addtemp(Src, addtemp(Dst, []));
	#dec{dst=Dst} -> addtemp(Dst, []);
	#inc{dst=Dst} -> addtemp(Dst, []);
	#jmp_fun{'fun'=Fun} -> addtemp(Fun, []);
	#jmp_switch{temp=Temp} -> [Temp];
	#lea{mem=Mem} -> addtemp(Mem, []);
	#move{src=Src,dst=Dst} -> addtemp(Src, dst_use(Dst));
	#pseudo_call{'fun'=Fun, args=Args} -> addtemp(Fun, addtemps(Args, []));
	#pseudo_return{args=Args} -> addtemps(Args, []);
	#pseudo_tailcall{'fun'=Fun,args=Args} -> addtemp(Fun, addtemps(Args, []));
	#push{src=Src} -> addtemp(Src, []);
	%% comment, jcc, jmp_label, label, nop, pseudo_jcc, ret
	_ -> []
    end.

dst_use(Dst) ->
    case Dst of
	#x86_mem{base=Base,off=Off} -> addtemp(Base, addtemp(Off, []));
	_ -> []
    end.

%%%
%%% src_use(Src) -- Return set of temps used by a source operand.
%%%

src_use(Src) ->
    addtemp(Src, []).

%%%
%%% Auxiliary operations on sets of temps
%%%

addtemps([Arg|Args], Set) ->
    addtemps(Args, addtemp(Arg, Set));
addtemps([], Set) ->
    Set.

addtemp(Arg, Set) ->
    case Arg of
	#x86_temp{} -> add(Arg, Set);
	#x86_mem{base=Base,off=Off} -> addtemp(Off, addtemp(Base, Set));
	_ -> Set
    end.

add(Arg, Set) ->
    case lists:member(Arg, Set) of
	false -> [Arg|Set];
	_ -> Set
    end.
