%%%----------------------------------------------------------------------
%%% File    : otp_4013.erl
%%% Author  : Micael Karlberg <Micael.Karlberg@ericsson.com>
%%% Purpose : Test and verify the ticket OTP-4013
%%% Created : 30 May 2001 by Micael Karlberg <Micael.Karlberg@ericsson.com>
%%%----------------------------------------------------------------------

%% otp_4013:do6().
%% otp_4013:do().

-module(otp_4013).
-author('Micael.Karlberg@ericsson.com').

-export([do/0,do6/0]).

do6() ->
    do1(msg6()).

do() ->
    [do(M) || M <- msgs()].

do({No,M,_}) when No == 6 ->
    i("DO TEST ~p",[No]),
    %% put(dbg,true),
    Res = do1(M),
    %% erase(dbg),
    Res;
do({No,M,_}) ->
    i("DO TEST ~p",[No]),
    do1(M).

do1(M) ->
    MBin = list_to_binary(M),
    case megaco_compact_text_encoder:decode_message([],MBin) of
	{ok,Message} ->
	    i("message decoded: ~n~p",[Message]),
	    ok;
	{error,[{reason,Reason,Line},{token,Tokens},{chars,Chars}]} ->
	    i("encoding failed:"
	      "~n   Reason: ~p"
	      "~n   Line:   ~p"
	      "~n   Tokens: ~p"
	      "~n   Chars:  ~p",[Reason,Line,Tokens,Chars]),
	    error;
	{error,[{reason,Reason},{token,Tokens},{chars,Chars}]} ->
	    i("encoding failed:"
	      "~n   Reason: ~p"
	      "~n   Tokens: ~p"
	      "~n   Chars:  ~p",[Reason,Tokens,Chars]),
	    error;
	{error,Reason} ->
	    i("encoding failed: ~p",[Reason]),
	    error
    end.


i(F,A) ->
    io:format("*** INFO: " ++ F ++ "~n",A).

msgs() -> 
    [
     {1,msg1(),error},
     {2,msg2(),ok},
     {3,msg3(),ok},
     {4,msg4(),ok},
     {5,msg5(),ok},
     {6,msg6(),ok},
     {7,msg7(),ok},
     {8,msg8(),ok}
    ].
    
msg1() ->
    "MEGCAO/1 MG1 T=12345678{C=-{SC=root{SV{MT=RS,RE=901}}}}".

msg2() ->
    "MEGACO/1 MG1 T=12345678{C=-{SC=root{SV{MT=RS,RE=901}}}}".

msg3() ->
    "!/1 MG1 T=12345678{C=-{SC=root{SV{MT=RS,RE=901}}}}".

msg4() ->
	"!/1 ML T=233350{C=3554454{A=stedevice/01{M{O{MO=SR,RV=OFF,RG=OFF,tdmc/ec={1,0}}}}}}".
    
msg5() ->
	"!/1 ML T=233350{C=3554454{A=stedevice/01{M{O{MO=SR,RV=OFF,RG=OFF,tdmc/ec=[1,0]}}}}}". 

msg6() ->
	"!/1 ML T=233350{C=3554454{A=stedevice/01{M{O{MO=SR,RV=OFF,RG=OFF,tdmc/ec=[1:0]}}}}}". 

msg7() ->
	"!/1 ML T=233350{C=3554454{A=stedevice/01{M{O{MO=SR,RV=OFF,RG=OFF,tdmc/ec={3,2,1,0}}}}}}".
    
msg8() ->
	"!/1 ML T=233350{C=3554454{A=stedevice/01{M{O{MO=SR,RV=OFF,RG=OFF,tdmc/ec=[3,2,1,0]}}}}}". 

